﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;

namespace HeuristicLab.Analysis.FitnessLandscape {
  [Item("Up/DownSelector", "A selection operator that moves towards the next local optimum, when found reverses direction and so on.")]
  [StorableType("67964036-8CE4-4987-9495-8174D1B218A5")]
  public class UpDownSelector : InstrumentedOperator, IStochasticOperator {

    #region Parameters
    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public ILookupParameter<BoolValue> MoveTowardsOptimumParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["MoveTowardsOptimum"]; }
    }
    public ILookupParameter<DoubleValue> BaseQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["BaseQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    #endregion

    #region Construction & Cloning

    [StorableConstructor]
    protected UpDownSelector(StorableConstructorFlag _) : base(_) { }
    protected UpDownSelector(UpDownSelector original, Cloner cloner) : base(original, cloner) { }
    public UpDownSelector() {
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", "Whether the problem is a maximization or minimization problem."));
      Parameters.Add(new LookupParameter<BoolValue>("MoveTowardsOptimum", "Specifies whether the selector should optimize towards or away from the optimum."));
      Parameters.Add(new LookupParameter<DoubleValue>("BaseQuality", "The base quality value to compare to. This is required to determine wheter a local optimum has been found and the direction should be reversed."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "The qualities of the solutions."));
      Parameters.Add(new LookupParameter<IRandom>("Random", "Random number generator"));
      BaseQualityParameter.ActualName = "Quality";
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new UpDownSelector(this, cloner);
    }
    #endregion

    public override IOperation InstrumentedApply() {
      var scopes = ExecutionContext.Scope.SubScopes.ToList();
      var selected = Select(scopes);
      scopes.Remove(selected);
      ExecutionContext.Scope.SubScopes.Clear();
      ExecutionContext.Scope.SubScopes.Add(new Scope("Remaining"));
      ExecutionContext.Scope.SubScopes[0].SubScopes.AddRange(scopes);
      ExecutionContext.Scope.SubScopes.Add(new Scope("Selected"));
      ExecutionContext.Scope.SubScopes[1].SubScopes.Add(selected);
      return base.InstrumentedApply();
    }

    private IScope Select(List<IScope> scopes) {
      var maximization = MaximizationParameter.ActualValue.Value;
      bool optimize = true;
      if (MoveTowardsOptimumParameter.ActualValue == null) {
        MoveTowardsOptimumParameter.ActualValue = new BoolValue(true);
      } else {
        optimize = MoveTowardsOptimumParameter.ActualValue.Value;
      }
      var qualities = QualityParameter.ActualValue;
      var list = qualities.Select((x, i) => new { Index = i, Quality = x.Value }).ToList();
      var baseQuality = BaseQualityParameter.ActualValue.Value;
      if (double.IsNaN(baseQuality)) {
        baseQuality = maximization ? list.Max(x => x.Quality) : list.Min(x => x.Quality);
      }
      var random = RandomParameter.ActualValue;

      if (random != null)
        list.Shuffle(random);
      if (maximization && optimize || !maximization && !optimize) {
        list = list.OrderByDescending(x => x.Quality).ToList();
        if (list.Count > 0 && list[0].Quality <= baseQuality)
          MoveTowardsOptimumParameter.ActualValue = new BoolValue(!optimize);
      } else {
        list = list.OrderBy(x => x.Quality).ToList();
        if (list.Count > 0 && list[0].Quality >= baseQuality)
          MoveTowardsOptimumParameter.ActualValue = new BoolValue(!optimize);
      }
      return scopes[list[0].Index];
    }
  }
}
